#include "fft.hh"
#include <pthread.h>

bool infoFFT (fft_support_type supp)
{
   switch (supp) {
      case kFFTW:
#ifdef FFTW3         
         return true;
#else
         return false;
#endif
      case kFFTDouble:
#ifdef FFT_DOUBLE         
         return true;
#else
         return false;
#endif
      default:
         return false;
   }
}

static pthread_mutex_t fftw_mux = PTHREAD_MUTEX_INITIALIZER;

#ifdef FFTW3
//==========================================================================
//=                                                                        =
//=                                 Use FFTW3                              =
//=                                                                        =
//==========================================================================
#include <complex>
#include "fftw3.h"

//======================================  Complex fft
void 
wfft(fComplex x[], unsigned int N, int direc) {
    fftwf_complex* fx = reinterpret_cast<fftwf_complex*>(x);
    fftwf_plan p;
    pthread_mutex_lock (&fftw_mux);
    if (direc > 0) {
        p = fftwf_plan_dft_1d(N, fx, fx, FFTW_FORWARD,  FFTW_ESTIMATE);
    } else {
        p = fftwf_plan_dft_1d(N, fx, fx, FFTW_BACKWARD, FFTW_ESTIMATE);
    }
    pthread_mutex_unlock (&fftw_mux);
    fftwf_execute(p);
    pthread_mutex_lock (&fftw_mux);
    fftwf_destroy_plan(p);
    pthread_mutex_unlock (&fftw_mux);
}

//======================================  Real fft
void 
wfft(const float x[], fComplex y[], unsigned int N) {
    fftwf_complex* fy = reinterpret_cast<fftwf_complex*>(y);
    pthread_mutex_lock (&fftw_mux);
    fftwf_plan p = fftwf_plan_dft_r2c_1d(N, 
					 const_cast<float*>(x), 
					 fy, 
					 FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
    pthread_mutex_unlock (&fftw_mux);
    fftwf_execute(p);
    pthread_mutex_lock (&fftw_mux);
    fftwf_destroy_plan(p);
    pthread_mutex_unlock (&fftw_mux);
}


//======================================  Real inverse fft
void 
wfft(const fComplex y[], float x[], unsigned int N) {
    const fftwf_complex* fy = reinterpret_cast<const fftwf_complex*>(y);
    pthread_mutex_lock (&fftw_mux);
    fftwf_plan p = fftwf_plan_dft_c2r_1d(N, 
					 const_cast<fftwf_complex*>(fy), 
					 x, 
					 FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
    pthread_mutex_unlock (&fftw_mux);
    fftwf_execute(p);
    pthread_mutex_lock (&fftw_mux);
    fftwf_destroy_plan(p);
    pthread_mutex_unlock (&fftw_mux);
}

#ifdef FFT_DOUBLE
//======================================  Complex fft (double)
void 
wfft(dComplex x[], unsigned int N, int direc) {
    fftw_complex* fx = reinterpret_cast<fftw_complex*>(x);
    fftw_plan p;
    pthread_mutex_lock (&fftw_mux);
    if (direc > 0) {
        p = fftw_plan_dft_1d(N, fx, fx, FFTW_FORWARD,  FFTW_ESTIMATE);
    } else {
        p = fftw_plan_dft_1d(N, fx, fx, FFTW_BACKWARD, FFTW_ESTIMATE);
    }
    pthread_mutex_unlock (&fftw_mux);
    fftw_execute(p);
    pthread_mutex_lock (&fftw_mux);
    fftw_destroy_plan(p);
    pthread_mutex_unlock (&fftw_mux);
}

//======================================  Real fft (double)
void 
wfft(const double x[], dComplex y[], unsigned int N) {
    fftw_complex* fy = reinterpret_cast<fftw_complex*>(y);
    pthread_mutex_lock (&fftw_mux);
    fftw_plan p = fftw_plan_dft_r2c_1d(N, 
				       const_cast<double*>(x), 
				       fy, 
				       FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
    pthread_mutex_unlock (&fftw_mux);
    fftw_execute(p);
    pthread_mutex_lock (&fftw_mux);
    fftw_destroy_plan(p);
    pthread_mutex_unlock (&fftw_mux);
}


//======================================  Real inverse fft (double)
void 
wfft(const dComplex y[], double x[], unsigned int N) {
    const fftw_complex* fy = reinterpret_cast<const fftw_complex*>(y);
    pthread_mutex_lock (&fftw_mux);
    fftw_plan p = fftw_plan_dft_c2r_1d(N, 
				       const_cast<fftw_complex*>(fy), 
				       x, 
				       FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
    pthread_mutex_unlock (&fftw_mux);
    fftw_execute(p);
    pthread_mutex_lock (&fftw_mux);
    fftw_destroy_plan(p);
    pthread_mutex_unlock (&fftw_mux);
}
#endif

#else
//==========================================================================
//=                                                                        =
//=                                 Use FFTPAK                             =
//=                                                                        =
//==========================================================================

#include "fftpak.hh"

//======================================  FFT plan
static fftplan plan;

//======================================  Complex fft
void wfft(fComplex x[], unsigned int N, int direc) {
    if (N != plan.c.LastN) {
        if (plan.c.NAlloc < N) {
	    if (plan.c.WSave) delete[] plan.c.WSave;
	    plan.c.WSave = new FLOAT[4*N+15];
	    plan.c.NAlloc = N;
	}
	cffti(N, plan.c.WSave);
	plan.c.LastN = N;
    }
    if (direc > 0) {
        cfftf(N, (FLOAT*)x, plan.c.WSave);
    } else {
        cfftb(N, (FLOAT*)x, plan.c.WSave);
    }
}

//======================================  Real fft
void wfft(const float x[], fComplex y[], unsigned int N) {
    if (N != plan.r.LastN) {
        if (plan.r.NAlloc < N) {
	    if (plan.r.WSave) delete[] plan.r.WSave;
	    plan.r.WSave = new FLOAT[2*N+15];
	    plan.r.NAlloc = N;
	}
	rffti(N, plan.r.WSave);
	plan.r.LastN = N;    
    }
    rfftf(N, x, y, plan.r.WSave);
}

//======================================  Real inverse fft
void wfft(const fComplex y[], float x[], unsigned int N) {
   printf ("Inverse FFT not supported\n");
}

#endif

#ifndef FFT_DOUBLE
void wfft(dComplex x[], unsigned int N, int direc) {
   printf ("Double precision FFTs not supported\n");
}

void wfft(const double x[], dComplex y[], unsigned int N) {
   printf ("Double precision FFTs not supported\n");
}

void wfft(const dComplex y[], double x[], unsigned int N) {
   printf ("Double precision FFTs not supported\n");
}
#endif
