/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TRIGSTREAM_HH
#define TRIGSTREAM_HH

#include "Time.hh"
#include "TrigBase.hh"
#include "Segment.hh"
#include "Process.hh"
#include "LdasDBWriter.hh"
#include "DQSegWriter.hh"
#include <string>
#include <list>

#define defTableCount 100

std::string getUnique(void);

/**  @name Trigger stream
  *  The trigger stream class is used by the trigger manager class to output
  *  trigger or segment table rows to the specified datatbase,
  *  @memo Trigger stream output class
  *  @author J. Zweizig
  *  @version 1.0; Modified July 6, 2005
  */
class TrigStream {
public:
    typedef std::list<trig::TrigBase> TriggerDB;
    typedef TriggerDB::iterator trig_iter;

    typedef std::list<trig::Segment> SegmentDB;
    typedef SegmentDB::iterator segm_iter;

    typedef unsigned long mask_type;
    typedef unsigned long count_type;

    struct param {
	param(void);
	Interval    _Inactive;
	Interval    _MaxDelay;
	count_type  _TableCount;
	std::string _OutDir;
    };

    struct stream_stats {
	stream_stats(void);
	void count_seg(void);
	void count_trig(void);
	void count_write(const std::string& file);
	count_type  errTot;
	count_type  fileTot;
	count_type  segTot;
	count_type  trigTot;
	std::string lastFile;
	Time        lastUpdate;
	Time        lastWrite;
    };

public:
    TrigStream(const std::string& name, mask_type mask);
    virtual ~TrigStream(void);
    virtual void addProcess(const Process& pdb);
    virtual bool addSegment(const trig::Segment& seg, mask_type dest);
    virtual bool addTrigger(const trig::TrigBase& tr, mask_type dest);
    virtual TrigStream* clone(void) const=0;
    virtual bool empty(void) const;
    int getDebug(void) const;
    Interval getTimeOut(const Time& now) const;
    Interval getInactive(void) const;
    Interval getMaxDelay(void) const;
    virtual const char* getName(void) const;
    virtual count_type getNTrigs(void) const=0;
    virtual count_type getNSegs(void) const=0;
    virtual const char* getType(void) const=0;
    const param& refParams(void) const;
    const ProcessDB& refProcDB(void) const;
    const stream_stats& refStats(void) const;
    void removeProc(const std::string& proc);
    void setDebug(int lvl);
    void setDestMsk(mask_type mask);
    void setExcludeMsk(mask_type mask);
    void setInactive(Interval dt);
    void setMaxDelay(Interval dt);
    void setOutDir(const std::string& dir);
    void setParam(const param& par);
    bool testWrite(const trig::Segment& seg, mask_type dest) const;
    bool testWrite(const trig::TrigBase& tr, mask_type dest) const;
    bool writeNow(bool force) const;
    virtual void WriteDB(bool force);
    virtual void PurgeProcDB(void);
    void tick(const Time& now);

protected:
    void setTimer(const Time& t);

private:
    std::string  mName;
    int          mDebug;
    mask_type    mDestMsk;
    mask_type    mExcludeMsk;

protected:
    param        mParam;
    ProcessDB    mProcDB;
    stream_stats mStats;
};

inline int 
TrigStream::getDebug(void) const {
    return mDebug;
}

inline Interval 
TrigStream::getInactive(void) const {
    return mParam._Inactive;
}

inline Interval 
TrigStream::getMaxDelay(void) const {
    return mParam._MaxDelay;
}

inline const char*
TrigStream::getName(void) const {
    return mName.c_str();
}

inline const TrigStream::param& 
TrigStream::refParams(void) const {
    return mParam;
}

inline const ProcessDB& 
TrigStream::refProcDB(void) const {
    return mProcDB;
}

inline const TrigStream::stream_stats&
TrigStream::refStats(void) const {
    return mStats;
}


//======================================  Ldas database stream
/**  @name LDAS stream
  *  The LDAS stream class is used by the trigger manager class to output
  *  trigger or segment table rows to an Ldas database,
  *  @memo LDAS stream output class
  *  @author J. Zweizig
  *  @version 1.0; Modified July 6, 2005
  */
class LdasStream : public TrigStream {
public:
    typedef TrigStream::count_type count_type;
    LdasStream(const std::string& name, mask_type mask);
    ~LdasStream(void);
    virtual LdasStream* clone(void) const;
    bool addSegment(const trig::Segment& seg, mask_type dest);
    bool addTrigger(const trig::TrigBase& tr, mask_type dest);
    bool empty(void) const;
    count_type getNTrigs(void) const;
    count_type getNSegs(void)  const;
    const char* getType(void) const {return "LDAS";}
    void WriteDB(bool force);
private:
    trig::LdasDBWriter mWriter;
};

//======================================  Segment database stream
/**  @name Segment stream
  *  The Segment stream class is used by the trigger manager class to output
  *  trigger or segment table rows to a Segment database,
  *  @memo Segment stream output class
  *  @author J. Zweizig
  *  @version 1.0; Modified July 6, 2005
  */
class SegDBStream : public TrigStream {
public:
    typedef TrigStream::count_type count_type;
    SegDBStream(const std::string& name, mask_type mask, const std::string& dbname);
    ~SegDBStream(void);
    virtual SegDBStream* clone(void) const;
    bool addSegment(const trig::Segment& seg, mask_type dest);
    bool empty(void) const;
    count_type getNTrigs(void) const;
    count_type getNSegs(void) const;
    const char* getType(void) const {return "Segment";}
    void WriteDB(bool force);
private:
    trig::DQSegWriter mWriter;
};

//======================================  Epics stream
/**  @name Epics stream
  *  The Epics stream class is used by the trigger manager class to output
  *  triggers to an Epics data server,
  *  @memo Epics stream output class
  *  @author J. Zweizig
  *  @version 1.0; Modified July 6, 2005
  */
class EpxStream : public TrigStream {
public:
    typedef TrigStream::count_type count_type;
    using TrigStream::addTrigger;
    EpxStream(const std::string& name, mask_type mask);
    ~EpxStream(void);
    virtual EpxStream* clone(void) const;
    count_type getNTrigs(void) const;
    count_type getNSegs(void)  const;
    const char* getType(void) const {return "Epics";}
    bool addTrigger(const trig::TrigBase& trg, mask_type dest);
};

#endif  //  TRIGSTREAM_HH
