#ifndef _LIGO_FILTERPARSE_H
#define _LIGO_FILTERPARSE_H
/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: FilterParse						*/
/*                                                         		*/
/* Module Description: parse a filter					*/
/*                                                         		*/
/* Revision History:					   		*/
/* Rel   Date     Programmer  	Comments				*/
/* 0.1	 17Jul02  D. Sigg    	First release		   		*/
/*                                                         		*/
/* Documentation References:						*/
/*	Man Pages: FilterParse.html					*/
/*	References: none						*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Daniel Sigg   (509) 372-8132  (509) 372-8137  sigg_d@ligo.mit.edu	*/
/*                                                         		*/
/*                                                         		*/
/*                      -------------------                             */
/*                                                         		*/
/*                             LIGO					*/
/*                                                         		*/
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.	*/
/*                                                         		*/
/*                     (C) The LIGO Project, 1999.			*/
/*                                                         		*/
/*                                                         		*/
/* Caltech				MIT		   		*/
/* LIGO Project MS 51-33		LIGO Project NW-17 161		*/
/* Pasadena CA 91125			Cambridge MA 01239 		*/
/*                                                         		*/
/* LIGO Hanford Observatory		LIGO Livingston Observatory	*/
/* P.O. Box 1970 S9-02			19100 LIGO Lane Rd.		*/
/* Richland WA 99352			Livingston, LA 70754		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#include "Complex.hh"
#include "FilterBase.hh"
#include <string>
#include <vector>


/** Parse a filter.
    This class parses a filter. For a detailed description see 
    the FilterDesign class.

    @memo Filter parsing
    @author Written July 2002 by Masahiro Ito
    @version 1.0
 ************************************************************************/
   class FilterParse {
   public:
      typedef std::vector< std::string > StrList;
   
      /** Constructs a filter parser. 
          @memo Constructor.
          @param spec Filter design specification
       ******************************************************************/
      explicit FilterParse (const char* spec = 0);
      /** Destructs the filter design class.
          @memo Destructor.
       ******************************************************************/
      virtual ~FilterParse() {
      }
      /** Parse a filter and call the corresponding methods.
          @memo Parse a filter
          @param spec New filter specification
          @return true if successful
       ******************************************************************/
      virtual bool filter (const char* formula);
   
      /// Gain
      virtual bool gain (double g, const char* format = "scalar") {
         return false; }
      /// Pole
      virtual bool pole (double f, double gain, const char* plane = "s") {
         return false; }
      virtual bool pole (double f, const char* plane = "s") {
         return false; }
      /// Zero
      virtual bool zero (double f, double gain, const char* plane = "s") {
      
         return false; }
      virtual bool zero (double f, const char* plane = "s") {
      
         return false; }
      /// Pole2
      virtual bool pole2 (double f, double Q, double gain,
                        const char* plane = "s") {
      
         return false; }
      virtual bool pole2 (double f, double Q, const char* plane = "s") {
         return false; }
      /// Zero2
      virtual bool zero2 (double f, double Q, double gain,
                        const char* plane = "s") {
         return false; }
      virtual bool zero2 (double f, double Q, const char* plane = "s") {
         return false; }
      /// ZPK
      virtual bool zpk (int nzeros, const dComplex* zero, 
                       int npoles, const dComplex* pole, double gain,
                       const char* plane = "s") {
         return false; }
      /// rpoly
      virtual bool rpoly (int nnumer, const double* numer, 
                        int ndenom, const double* denom, double gain) {
         return false; }
      /// Biquad
      virtual bool biquad (double b0, double b1, double b2, 
                        double a1, double a2) {
         return false; }
      /// SOS
      virtual bool sos (int nba, const double* ba, const char* format = "s") {
         return false; }
      /// z roots
      virtual bool zroots (int nzeros, const dComplex* zero,
                        int npoles, const dComplex* pole, double gain = 1.0) {
         return false; }
     /// Direct form
      virtual bool direct (int nb, const double* b, 
                        int na, const double* a) {
         return false; }
     /// Elliptic
      virtual bool ellip ( Filter_Type type, int order, double rp, double as, 
                        double f1, double f2 = 0.0 ) {
         return false; }
      /// Chebyshev type 1
      virtual bool cheby1 ( Filter_Type type, int order, double rp, 
                        double f1, double f2 = 0.0 ) {
         return false; }
      /// Chebyshev type 2
      virtual bool cheby2 ( Filter_Type type, int order, double as, 
                        double f1, double f2 = 0.0 ) {
         return false; }
      /// Butterworth
      virtual bool butter ( Filter_Type type, int order, double f1, 
                        double f2 = 0.0 ) {
         return false; }
      /// notch
      virtual bool notch ( double f0, double Q, double depth = 0.0 ) {
         return false; }
      /// resgain
      virtual bool resgain ( double f0, double Q, double height = 30.0 ) {
         return false; }
      /// comb
      virtual bool comb ( double f0, double Q, double amp = 0.0, int N = 0 ) {
         return false; }
      /// remez
      virtual bool remez (int N, int nBand, const double* Bands, 
                        const double* Func, const double* Weight = 0) {
         return false; }
      /// firw
      virtual bool firw (int N, Filter_Type type, const char* window, 
                        double Flow, double Fhigh = 0, double Ripple = 0, 
                        double dF = 0) {
         return false; }
      /// difference
      virtual bool difference() {
         return false; }
      /// decimateBy2
      virtual bool decimateBy2 (int N, int FilterID = 1) {
         return false; }
      /// linefilter
      virtual bool linefilter (double f, double T = 0.0, int fid = 1, 
                        int nT = 1) {
         return false; }
      /// limiter
      virtual bool limiter (const char* type, double l1, double l2 = 0, 
                        double l3 = 0) {
         return false; }
      /// multirate
      virtual bool multirate (const char* type, double m1, double m2 = 1E-3, 
                        double atten = 80) {
         return false; }
      /// mixer
      virtual bool mixer (double fmix, double phase = 0) {
         return false; }
      /// front end
      virtual bool frontend (const char* file, const char* module, 
                        int section = -1) {
         return false; }
      /// setgain
      virtual bool setgain (double f, double gain) {
         return false; }
      /// close loop
      virtual bool closeloop (double k = 1.0) {
         return false; }
   
      /// Split a line
      static void splitLine (const std::string& str_in, 
                        StrList& strlist, char delim = ','); 
      /// Remove quotes
      static bool removequote (const std::string& str_in, 
                        std::string& str_out);
      /// String to complex
      static bool str2cmplx (const std::string& str, dComplex& value); 
      /// get complex list
      static bool getComplexList (const std::string& value, 
                        std::vector<dComplex>& cmplxlist); 
      /// get real list
      static bool getDoubleList (const std::string& value, 
                        std::vector<double>& dbllist);
      /// get filter type
      static bool getFilterType (const char* type_name, 
                        Filter_Type& type);
      /// get filter string
      static std::string getFilterString (Filter_Type type);
   
   protected:
      /// add a filter ( name : filter name, arg : corresponding arguments ).
      virtual bool addfilter ( const std::string& name, 
                        const std::string& arg );
   };


#endif // _LIGO_FILTERPARSE_H
