////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                                LPEFilter.cc                                //
//                                                                            //
//               Linear predictive error filter implementation                //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#include "LPEFilter.hh"
#include "DVecType.hh"
#include <iostream>

//--------------------------------------  Default constructor.
LPEFilter::LPEFilter(int filterLength, int trainPeriod, int trainLength)
  : FIRFilter(), trainPeriod(trainPeriod), trainLength(trainLength),
    trainTime(0)
{
  setLength(filterLength);
}

//--------------------------------------  Copy constructor.
LPEFilter::LPEFilter(const LPEFilter& lpef)
  : FIRFilter(lpef), trainPeriod(lpef.trainPeriod),
    trainLength(lpef.trainLength), trainTime(0)
{
  reset();
}

//--------------------------------------  Destructor.
LPEFilter::~LPEFilter()
{
}

//--------------------------------------  Assignment operator
LPEFilter& LPEFilter::operator=(const LPEFilter& lpef)
{

  if (this != &lpef) {
    this->FIRFilter::operator=(lpef);
    trainPeriod = lpef.trainPeriod;
    trainLength = lpef.trainLength;
    trainTime = Time(0);
    reset();
  }

  return *this;

}

//--------------------------------------  Apply filter to TSeries
int LPEFilter::apply(const TSeries& in, TSeries& out)
{

  int trainFlag = 0;

  if (!trainTime) {
    trainFlag = !train(in);
  }

  if (trainPeriod != 0) {
    if (in.getStartTime() >= trainTime + trainPeriod) {
      trainFlag = !train(in);
    }
  }

  FIRFilter::apply(in, out);

  return trainFlag;

}

//--------------------------------------  Determine filter coefficients
int LPEFilter::train(const TSeries& data)
{

  if (!getLength()) {
    std::cerr << "LPEFilter: filter length not defined" << std::endl;
    return -1;
  }

  if (!trainLength) {
    std::cerr << "LPEFilter: training length not defined" << std::endl;
    return -1;
  }

  if (trainLength < getLength()) {
    std::cerr << "LPEFilter: insufficient training length" << std::endl;
    return -1;
  }

  if ((int)(data.getNSample()) < trainLength) {
    std::cerr << "LPEFilter: insufficient training data" << std::endl;
    return -1;
  }

  double* x = new double[trainLength];
  data.getData(trainLength, x);

  double* r = new double[getLength()];
  autocorr(x, r);

  double* c = new double[getLength()];
  levinson(r, c);

  setCoefs(c);

  setRate(1.0 / data.getTStep());

  delete [] c;
  delete [] r;
  delete [] x;

  trainTime = data.getStartTime();

  return 0;

}

//--------------------------------------  Compute autocorrelation vector
void LPEFilter::autocorr(double* x, double* r) const
{

  int lag;
  int sample;

  for (lag = 0; lag < getLength(); lag++) {
    r[lag] = 0.0;
    for (sample = lag; sample < trainLength; sample++) {
      r[lag] += x[sample] * x[sample - lag];
    }
    r[lag] /= trainLength;
  }

}

//--------------------------------------  Solve symmetric Yule-Walker problem
void LPEFilter::levinson(double* r, double* c) const
{

  int n, m;
  double tmp;
  double num, den;

  int M = getLength() - 1;

  c[1] = - r[1] / r[0];

  for (m = 1; m < M; m++) {

    num = r[m + 1];
    den = r[0];

    for (n = 1; n <= m; n++) {
      num += c[n] * r[m + 1 - n];
      den += c[n] * r[n];
    }

    c[m + 1] = - num / den;

    for (n = 1; n <= ((m + 1) >> 1); n++) {
      tmp = c[n] + c[m + 1] * c[m + 1 - n];
      c[m + 1 - n] = c[m + 1 - n] + c[m + 1] * c[n];
      c[n] = tmp;
    }

  }

  c[0] = 1;

}

//--------------------------------------  Set Train Period
void LPEFilter::setTrainPeriod(int period)
{
  trainPeriod = period;
}

//--------------------------------------  Set Train Length
void LPEFilter::setTrainLength(int length)
{
  trainLength = length;
}

//--------------------------------------  Set History
void LPEFilter::setHistory(const TSeries& in)
{

  reset();

  if (getRate() == 0.0) {
    setRate(1.0 / in.getTStep());
  }

  FIRFilter::apply(in);

}

//--------------------------------------  Get Train Period
int LPEFilter::getTrainPeriod(void) const
{
  return trainPeriod;
}

//--------------------------------------  Get Train Length
int LPEFilter::getTrainLength(void) const
{
  return trainLength;
}

//--------------------------------------  Get Train Time
Time LPEFilter::getTrainTime(void) const
{
  return trainTime;
}

//--------------------------------------  Get Transient Time
Interval LPEFilter::getTransientTime(void) const
{
  return getLength() / getRate();
}
