/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "ParseLine.hh"
#include "SegList.hh"
#include <sys/times.h>
#include <string>
#include <fstream>
#include <sstream>
#include <iostream>
#include <stdexcept>
#include <map>

#include "calc_engine/engine.hh"
#include "seg_stack.hh"

class SegEval : public Engine {
public:
    typedef std::map<std::string, LockSegList> dict_type;
public:
    SegEval(dict_type& t, const std::string& cmd, int debug=0);
    ~SegEval(void);
    const LockSegList& eval(void);
    virtual void eval_funct(const std::string& sym, int N);
    virtual void push_symbol(const std::string& sym);

private:
    dict_type& mDictionary;
    seg_stack  mStack;
};

class seg_calc {
public:
    seg_calc(int argc, const char* argv[]);
    void doit(void);
    void exec_cmd(const ParseLine& pl);
    void exec_stream(std::istream& file);
    void read_segfile(const std::string& file, const std::string& name,
		      const std::string& fmt="", const std::string& prefix="",
		      const std::string& version="");
    void usertime(void) const;

public:
    typedef std::map<std::string, LockSegList> seg_map;
    typedef seg_map::iterator seg_iter;
    typedef seg_map::const_iterator const_seg_iter;

private:
    bool exists(const std::string&) const;
    seg_iter       find(const std::string&);
    const_seg_iter find(const std::string&) const;
    seg_iter       find_nothrow(const std::string&);
    const_seg_iter find_nothrow(const std::string&) const;

private:
    std::string mCmdFile;
    std::string mInitCmd;
    std::string mOutFile;
    seg_map mSegMap;
    int mDebug;
    bool mVerbose;
    mutable double tDelta;
};

using namespace std;

//======================================  Main function
int main(int argc, const char* argv[]) {
    try {
	seg_calc x(argc, argv);
	x.doit();
    } catch (exception& e) {
	cerr << "Caught exception: " << e.what() << endl;
    }
}

//======================================  Utility methods
inline bool
seg_calc::exists(const string& name) const {
    return mSegMap.find(name) != mSegMap.end();
}

inline seg_calc::seg_iter
seg_calc::find_nothrow(const string& name) {
    return mSegMap.find(name);
}

inline seg_calc::const_seg_iter
seg_calc::find_nothrow(const string& name) const {
    return mSegMap.find(name);
}

inline seg_calc::seg_iter
seg_calc::find(const string& name) {
    seg_iter p = mSegMap.find(name);
    if (p == mSegMap.end()) 
	throw runtime_error(string("Unknown segment list: ") + name);
    return p;
}

inline seg_calc::const_seg_iter
seg_calc::find(const string& name) const {
    const_seg_iter p = mSegMap.find(name);
    if (p == mSegMap.end()) 
	throw runtime_error(string("Unknown segment list: ") + name);
    return p;
}

void
seg_calc::usertime(void) const {
    struct tms t;
    times(&t);
    double tNow(t.tms_utime/double(sysconf(_SC_CLK_TCK)));
    cerr<< "user time: " << tNow-tDelta <<  " (total: " << tNow << ")" << endl;
    tDelta = tNow;
}

//======================================  Constructor
seg_calc::seg_calc(int argc, const char* argv[])
    : mDebug(0), mVerbose(false), tDelta(0)
{
    string fmt, pfx;

    //----------------------------------  Parse command string.
    for (int i=1; i<argc; ++i) {
	string argi = argv[i];
	if (argi == "-c") {
	    if (mInitCmd.empty()) {
		mInitCmd = argv[++i];
	    } else {
		mInitCmd += "; ";
		mInitCmd += argv[++i];
	    }
	} else if (argi == "-f") {
	    fmt = argv[++i];
	} else if (argi == "-i") {
	    mCmdFile = argv[++i];
	} else if (argi.substr(0,2) == "-s") {
	    string name;
	    if (argi.size() > 2) {
		if (argi[2] == ':') name = argi.substr(3);
		else throw runtime_error("Segment name syntax error");
	    }
	    read_segfile(argv[++i], name, fmt, pfx);
	} else if (argi == "-v" || argi == "--verbose") {
	    mVerbose = true;
	} else if (argi == "--debug") {
	    if (i+1 >= argc || *(argv[i+1]) == '-') mDebug = 1;
	    else                   mDebug = strtol(argv[++i], 0, 0);
	} else if (argi[0] != '-') {
	    read_segfile(argi, "", fmt, pfx);
	} else {
	    throw  runtime_error(string("Undefined keyword: ") + argv[i]);
	}
    }
}

//======================================  Execute command stream
void
seg_calc::doit(void) {

    //----------------------------------  Execute a list of commands
    if (!mInitCmd.empty()) {
	istringstream cmdin(mInitCmd);
	exec_stream(cmdin);
    }

    //----------------------------------  Execute a list of commands
    if (!mCmdFile.empty()) {
	if (mCmdFile == "-") {
	    exec_stream(cin);
	} else {
	    ifstream in(mCmdFile.c_str());
	    if (!in.is_open()) 
		throw runtime_error(string("Can't open file: ") + mCmdFile);
	    exec_stream(in);
	}
    }

}

//======================================  Command execution
void 
seg_calc::exec_stream(std::istream& file) {
    ParseLine pl(file);
    if (!pl.isOpen()) throw runtime_error("Can't open file.");
    if (mVerbose) pl.setLog(cerr);
    while (pl.getLine() >= 0) {
	if (pl.getCount() > 0) {
	    try {
		exec_cmd(pl);
	    } catch (exception& e) {
		cerr << "Caught exception: '" << e.what() << "' on line: "
		     << pl.getLineNumber() << endl;
	    }
	    if (mVerbose) usertime();
	}
    }
}

//======================================  Command execution
void 
seg_calc::exec_cmd(const ParseLine& pl) {
    int nArg = pl.getCount();
    string cmdi = pl[0];

    //------------------------------  Define a segment list
    if (cmdi == "define") {
	string eval_cmd, name, ifo("xx");
	bool replace=false, logit=false;
	int i0=1;
	if (*(pl[i0]) != '-' ) name = pl[i0++];
	for (int i=i0; i<nArg; ++i) {
	    string argi=pl[i];
	    if (argi == "-eval") {
		eval_cmd = pl[++i];
	    } else if (argi == "-ifo") {
		ifo = pl[++i];
	    } else if (argi == "-log") {
		logit = true;
	    } else if (argi == "-name") {
		name = pl[++i];
	    } else if (argi == "-replace") {
		if (i+1 < nArg && *(pl[i+1]) != '-') name = pl[++i];
		replace = true;
	    } else {
		throw runtime_error(string("Undefined argument: ")+argi);
	    }
	}
	LockSegList r(ifo.c_str());
	if (name.empty()) {
	    throw("Segment file not specified");
	} else if (!replace && exists(name)) {
	    throw runtime_error(string("Segment ") + name + " exists.");
	} else {
	    if (!eval_cmd.empty()) {
		SegEval ev(mSegMap, eval_cmd, mDebug);
		r = ev.eval();
	    }
	    seg_iter p = find_nothrow(name);
	    if (p != mSegMap.end()) p->second = r;
	    else              mSegMap.insert(seg_map::value_type(name, r));
	}
	if (logit) cout << "# " << name << " defined as " << eval_cmd << endl;
	if (mVerbose) cerr << "SegList: " << name << " defined with "
			   << r.size() << " segments, " << r.live()
			   << " seconds live." << endl;

    //----------------------------------  Echo a line
    } else if (cmdi == "echo") {
	cout << pl[1] << endl;

    //----------------------------------  Execute a command file.
    } else if (cmdi == "exec") {
	if (nArg < 2) 
	    throw runtime_error("exec must have a filename argument.");
	string wd;
	for (int i=1; i<nArg-1; ++i) {
	    string argi=pl[i];
	    if (argi == "-cwd") {
		wd = pl[++i];
	    } else {
		throw runtime_error(string("Undefined argument: ") + argi);
	    }
	}

	string macro = pl[nArg-1];
	if (macro[0] != '/' && !wd.empty()) macro = wd + "/" + macro;
	ifstream in(macro.c_str());
	if (!in.is_open()) throw runtime_error(string("Can't open: ")+macro);
	if (wd.empty()) {
	    exec_stream(in);
	} else {
	    char savewd[1024];
	    getcwd(savewd, sizeof(savewd));
	    try {
		chdir(wd.c_str());
		exec_stream(in);
		chdir(savewd);
	    } catch (exception& e) {
		chdir(savewd);
		throw e;
	    }
	}

    //----------------------------------  Print a segment index
    } else if (cmdi == "index") {
	string name;
	for (int i=1; i<nArg; ++i) {
	    string argi=pl[i];
	    if (argi == "-name") {
		name = pl[++i];
	    } else {
		throw runtime_error(string("Undefined argument: ") + argi);
	    }
	}

	cout << "Name  size  live" << endl;
	for (const_seg_iter i=mSegMap.begin(); i != mSegMap.end(); ++i) {
	    if (name.empty() || name == i->first) {
		LockSegList::size_type N=i->second.size();
		cout << i->first << "  " << N << "  " << long(double(i->second.live())+0.5);
		if (!N) {
		    cout << "       --         --     ";
		} else {
		    cout << "  " << i->second[0].getStartTime().getS() 
			 << "  " << i->second[N-1].getEndTime().getS();
		}
		cout << endl;
	    }
	}
	
    //----------------------------------  Read in a segment list
    } else if (cmdi == "read") {
	string name, sfile, fmt, prefix, version;
	for (int i=1; i<nArg; ++i) {
	    string argi=pl[i];
	    if (argi == "-file") {
		sfile = pl[++i];
	    } else if (argi == "-format") {
		fmt = pl[++i];
	    } else if (argi == "-name") {
		name = pl[++i];
	    } else if (argi == "-prefix") {
		prefix = pl[++i];
	    } else if (argi == "-version") {
		version = pl[++i];
	    } else {
		throw runtime_error(string("Undefined argument: ") + argi);
	    }
	}
	if (sfile.empty()) throw("Segment file not specified");
	read_segfile(sfile, name, fmt, prefix, version);

    //----------------------------------  Read in a segment list
    } else if (cmdi == "remove") {
	string name;
	for (int i=1; i<nArg; ++i) {
	    string argi=pl[i];
	    if (argi == "-name") {
		name = pl[++i];
	    } else {
		throw runtime_error(string("Undefined argument: ") + argi);
	    }
	}
	if (name.empty()) throw runtime_error("No segment list specified.");
	mSegMap.erase(find(name));
 
    //----------------------------------  Write a segment list
    } else if (cmdi == "write") {
	string name, sfile;
	string format("%i %s %d");
	for (int i=1; i<nArg; ++i) {
	    string argi=pl[i];
	    if (argi == "-name") {
		name = pl[++i];
	    } else if (argi == "-file") {
		sfile = pl[++i];
	    } else if (argi == "-format") {
		format = pl[++i];
	    } else {
		throw runtime_error(string("Undefined argument: ") + argi);
	    }
	}

	ostream* of=0;
	bool     alloc_stream=false;
	if (sfile.empty() || sfile == "-") {
	    of = &cout;
	} else {
	    of = new ofstream(sfile.c_str());
	    if (!dynamic_cast<ofstream*>(of)->is_open()) 
		throw runtime_error(string("Unable to open file:") + sfile);
	    alloc_stream = true;
	}
	for (const_seg_iter i=mSegMap.begin(); i!=mSegMap.end(); ++i) {
	    if (name.empty() || name == i->first) {
		int N=i->second.size();
		for (int inx=0; inx<N; ++inx) {
		    i->second[inx].write(*of, format) << endl;
		}
	    }
	}
	if (alloc_stream) delete of;

    //------------------------------  Write a segment list
    } else {
	throw runtime_error(string("Undefined command:") + cmdi);
    }
}

//======================================  Configuration
void 
seg_calc::read_segfile(const string& file, const string& name, 
		       const string& fmt, const string& prefix,
		       const string& version) {
    ParseLine pl(file.c_str());
    if (!pl.isOpen()) 
	throw runtime_error(string("Unable to open segment file: ")+file);

    seg_map  lclmap;
    seg_iter lclitr = lclmap.end();
    string   lclname;

    int iStart(-1), iName(-1), iEnd(-1), iDt(-1), iID(-1), iVsn(-1), iAct(-1);
    int nCol(1);

    //----------------------------------  Parse format specifier
    if (!fmt.empty()) {
	string pfmt(fmt);
	for (int i=0; !pfmt.empty(); ++i) {
	    switch (pfmt[0]) {
	    case 's':
		iStart = i;
		break;
	    case 'e':
		iEnd = i;
		break;
	    case 'd':
		iDt = i;
		break;
	    case 'n':
		iName = i;
		break;
	    case 'v':
		iVsn = i;
		break;
	    case 'a':
		iAct = i;
		break;
	    case 'i':
		iID = i;
		break;
	    default:
		break;
	    }
	    string::size_type n=pfmt.find(":");
	    if (n != string::npos) {
		pfmt.erase(0,pfmt.find(":")+1);
		nCol++;
	    } else {
		pfmt.clear();
	    }
	}

	if (mDebug) cerr << "Start/Name/End/Dt/ID=" << iStart << "/" << iName
			 << "/"  << iEnd << "/" << iDt << "/" << iID << endl;

	if (iStart < 0 || (iEnd < 0 && iDt < 0)) {
	    throw runtime_error(string("Format is insufficient: ")+fmt);
	}
    }

    //----------------------------------  Loop over segment lines.
    while (pl.getLine()>=0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	if (iStart < 0) {
	    if (mDebug) cerr << "Identifying fields in: " << endl;
	    for (int i=0; i<nArg; ++i) {
		if (mDebug) cerr << pl[i] << " ";
		if (string(pl[i]).find_first_not_of("0123456789") == string::npos) {
		    double t=pl.getDouble(i);
		    if (t>6e8 && t<2e9) {
			if (iStart < 0) iStart = i;
			else iEnd = i;
		    } else if (t>0 && iStart >= 0) {
			iDt = i;
		    } else {
			iID = i;
		    }
		} else {
		    iName = i;
		}
	    }
	    if (iStart<0 || (iEnd<0 && iDt<<0)) {
		throw runtime_error("Unable to identify segment list fields");
	    } else if (mDebug) {
		cerr << endl;
		cerr << "Start/Name/End/Dt/ID=" << iStart << "/" << iName
		     << "/"  << iEnd << "/" << iDt << "/" << iID << endl;
	    }
	    nCol = nArg;
	}

	//-----------------------------  Check that the line is wide enough
	if (nArg < nCol) {
	    cerr << "Line: " << pl.getLineNumber() << " is too short ("
		 << nArg << " words) for format." << endl;
	    continue;
	}

	//-----------------------------  check for active
	if (iAct >= 0 && pl.getInt(iAct) == 0) continue;

	//-----------------------------  Get the segment name
	string segname = prefix;
	if (iName < 0) {
	    segname += name;
	} else if (name.empty() || name == pl[iName]) {
	    segname += pl[iName];
	} else {
	    continue;
	}

	if (iVsn < 0) {
	    if (!version.empty()) {
		segname += "v";
		segname += version;
	    }
	} else if (version.empty() || version == pl[iVsn]) {
	    segname += "v";
	    segname += pl[iVsn];
	} else {
	    continue;
	}

	//-----------------------------  Get the segment time
	Time tStart(pl.getInt(iStart));
	Interval dT;
	if (iDt>=0) dT = pl.getDouble(iDt);
	else        dT = Time(pl.getInt(iEnd)) - tStart;
	int ID = iID<0 ? 0: pl.getInt(iID);
	LockSegment seg(ID, tStart, dT);

	//-----------------------------  Fill in the segment
	if (lclitr == lclmap.end() || segname != lclname) {
	    lclitr = lclmap.find(segname);
	    if (lclitr == lclmap.end()) {
		lclmap.insert(seg_map::value_type(segname, LockSegList(segname.c_str())));
		lclitr = lclmap.find(segname);
	    }
	    lclname = segname;
	}
	lclitr->second.stuff(seg);
    }

    //--------------------------------- Merge the files
    for (seg_iter p=lclmap.begin(); p != lclmap.end(); ++p) {
	p->second.sort();
	p->second.coalesce();
	if (mVerbose) cerr << "SegList: " << p->first << " defined with "
			   << p->second.size() << " segments, " 
			   << p->second.live() << " seconds live." << endl;

	seg_iter pm = find_nothrow(p->first);
	if (pm == mSegMap.end()) mSegMap.insert(*p);
	else                     pm->second |= p->second;
    }
}

//======================================  SegEval command constructor
SegEval::SegEval(dict_type& d, const std::string& cmd, int debug)
  : Engine(mStack), mDictionary(d)
{
    setDebug(debug);
    istringstream str(cmd);
    Parse(str);
}

//======================================  SegEval destructor
SegEval::~SegEval(void) {
}

//======================================  Evaluate a segment
const LockSegList& 
SegEval::eval(void) {
    crank();
    if (mStack.size() != 1) throw runtime_error("Invalid command string");
    return dynamic_cast<seg_token&>(mStack.top()).refSegment();
}

//======================================  Evaluate a function
void 
SegEval::eval_funct(const std::string& sym, int N) {
    if (sym == "not") {
        dynamic_cast<seg_token&>(mStack.top()).refSegment().invert();
    } else if (sym == "now") {
        mStack.push(double(Now().getS()));
    } else if (sym == "pad") {
	if (!mStack.numeric(1) || !mStack.numeric(2)) {
	    throw runtime_error("Pad sizes must be numeric");
	} else if (mStack.numeric(3)) {
	    throw runtime_error("Padded object must be segment list");
	} else {
	    int pEnd = int(mStack.pop_numeric());
	    int pBeg = int(mStack.pop_numeric());
	    mStack.top().refSegment().pad(pBeg, pEnd);
	}
    } else if (sym == "range") {
	if (mStack.numeric(1)) 
	    throw runtime_error("Operation not defined for numeric arguments");
	seg_token* t = dynamic_cast<seg_token*>(mStack.pop_token());
	LockSegList l("Xn");
	long nSegs=t->refSegment().size();
	if (nSegs != 0) {
	    l.insert(LockSegment(0, t->refSegment()[0].getStartTime(), 
				 t->refSegment()[nSegs-1].getEndTime()));
	}
	delete t;
	mStack.push(new seg_token(l));
    } else if (sym == "seg") {
        double end   = mStack.pop_numeric();
	double start = mStack.pop_numeric();
	LockSegList l("Xn");
	l.insert(LockSegment(0, Time(long(start)), end-start));
	mStack.push(new seg_token(l));
    } else if (sym == "xor") {
	if (mStack.numeric(1) || mStack.numeric(2)) {
	    throw runtime_error("Operation not defined for numeric arguments");
	} else {
	    seg_token* t = dynamic_cast<seg_token*>(mStack.pop_token());
	    dynamic_cast<seg_token&>(mStack.top()).refSegment() ^= 
		t->refSegment();
	    delete t;
	}
    } else {
        throw runtime_error(string("Undefined function:") + sym);
    }
}

//======================================  
void 
SegEval::push_symbol(const std::string& sym) {
    if (mDictionary.find(sym) == mDictionary.end()) {
        throw runtime_error(string("Undefined segment: ") + sym);
    }
    mStack.push(new seg_token(mDictionary[sym]));
}
